@@ -1,16 +1,15 @@
-import unittest
-
-import json
-
-import mock
+import unittest, mock
 
 from dynamodb_mapper.model import (ConnectionBorg, DynamoDBModel,
     autoincrement_int, MaxRetriesExceededError, MAX_RETRIES,
-    ConflictError, _python_to_dynamodb, _dynamodb_to_python, utc_tz,
-    _get_default_value, SchemaError, MAGIC_KEY, OverwriteError, InvalidRegionError,)
+    ConflictError, _python_to_dynamodb, _dynamodb_to_python, UTC, utc_tz,
+    SchemaError, MAGIC_KEY, OverwriteError, InvalidRegionError,)
 from boto.exception import DynamoDBResponseError
 from boto.dynamodb.exceptions import DynamoDBConditionalCheckFailedError
-import datetime
+from onctuous.validators import InRange, All, Length, Coerce
+from onctuous.errors import Invalid
+
+import json, datetime
 
 class mocked_region(object):
     def __init__(self, name):
@@ -164,9 +163,43 @@ class ToJsonTest(DynamoDBModel):
     }
 
 
+class SchemaValidators(DynamoDBModel):
+    __table__ = "SchemaValidators"
+    __hash_key__ = "name"
+    __schema__ = {
+        "name": All(Coerce(unicode), Length(min=3, max=15)),
+        "age": InRange(min=0),
+        "scores": [InRange(min=0, max=100)],
+    }
+
+class SchemaValidatorsBad(DynamoDBModel):
+    __table__ = "SchemaValidatorsBad"
+    __hash_key__ = "name"
+    __schema__ = {
+        "name": unicode,
+        "bad": set([1,2,3]),
+    }
+
 def return_42():
     return 42
 
+class SchemaCallableDefault(DynamoDBModel):
+    __table__ = "SchemaValidators"
+    __hash_key__ = "id"
+    __schema__ = {
+        "id": int,
+    }
+    __defaults__ = {"id": return_42}
+
+
+class TestUTC(unittest.TestCase):
+    def test_utcoffset(self):
+        self.assertEqual(datetime.timedelta(0), UTC().utcoffset(None))
+    def test_tzname(self):
+        self.assertEqual("UTC", UTC().tzname(None))
+    def test_dst(self):
+        self.assertEqual(datetime.timedelta(0), UTC().dst(None))
+
 
 class TestConnectionBorg(unittest.TestCase):
     def setUp(self):
@@ -175,6 +208,7 @@ class TestConnectionBorg(unittest.TestCase):
             "_aws_secret_access_key": None,
             "_region": None,
             "_connection": None,
+            "_tables_cache": {},
         }
 
     def tearDown(self):
@@ -183,6 +217,7 @@ class TestConnectionBorg(unittest.TestCase):
             "_aws_secret_access_key": None,
             "_region": None,
             "_connection": None,
+            "_tables_cache": {},
         }
 
     def test_borgness(self):
@@ -273,7 +308,6 @@ class TestConnectionBorg(unittest.TestCase):
         self.assertIs(None, borg1._region)
         self.assertIs(None, borg2._region)
 
-
     @mock.patch("dynamodb_mapper.model.boto")
     def test_set_credentials(self, m_boto):
         # Make sure internal state is set and shared
@@ -450,6 +484,26 @@ class TestConnectionBorg(unittest.TestCase):
 
         table.refresh.assert_called_once_with(wait_for_active=True)
 
+    @mock.patch("dynamodb_mapper.model.boto")
+    def test_create_table_validators(self, m_boto):
+        m_connection = m_boto.connect_dynamodb.return_value
+        m_create_schema = m_connection.create_schema
+        m_create_table = m_connection.create_table
+
+        table = ConnectionBorg().create_table(SchemaValidators, 10, 5, True)
+
+        m_create_schema.assert_called_once_with(
+            hash_key_name=SchemaValidators.__hash_key__,
+            hash_key_proto_value=u"",
+            range_key_name=None,
+            range_key_proto_value=None,
+        )
+
+        m_create_table.assert_called_once_with(
+            SchemaValidators.__table__, m_create_schema.return_value, 10, 5)
+
+        table.refresh.assert_called_once_with(wait_for_active=True)
+
 
 class TestTypeConversions(unittest.TestCase):
     def test_python_to_dynamodb_number(self):
@@ -475,8 +529,8 @@ class TestTypeConversions(unittest.TestCase):
 
     def test_dynamodb_to_python_unicode(self):
         self.assertEqual(u"hello", _dynamodb_to_python(unicode, u"hello"))
-        # Missing attributes become None
-        self.assertIsNone(_dynamodb_to_python(unicode, None))
+        # Missing attributes become neutral value
+        self.assertEqual(_dynamodb_to_python(unicode, None), u"")
 
     def test_python_to_dynamodb_set(self):
         self.assertEqual(set([1, 2, 3]), _python_to_dynamodb(set([1, 2, 3])))
@@ -491,8 +545,8 @@ class TestTypeConversions(unittest.TestCase):
         self.assertEqual(
             set(["foo", "bar", "baz"]),
             _dynamodb_to_python(set, set(["foo", "bar", "baz"])))
-        # Empty sets become None
-        self.assertIsNone(_dynamodb_to_python(set, None))
+        # Empty sets become... empty set
+        self.assertEqual(_dynamodb_to_python(set, None), set())
 
     def test_python_to_dynamodb_list(self):
         attacks = [
@@ -556,8 +610,10 @@ class TestTypeConversions(unittest.TestCase):
             _dynamodb_to_python(datetime.datetime,
                                 "2010-11-01T04:00:00.000013Z"))
 
-    def test_dynamodb_to_python_default(self):
-        self.assertIsNone(_dynamodb_to_python(datetime.datetime, None))
+    @mock.patch("dynamodb_mapper.model.datetime")
+    def test_dynamodb_to_python_default(self, m_datetime):
+        m_datetime.now.return_value = now = datetime.datetime.now(tz=utc_tz)
+        self.assertEqual(now, _dynamodb_to_python(m_datetime, None))
 
     def test_dynamodb_to_python_datetime_notz(self):
         # Timezone info is mandatory
@@ -576,47 +632,6 @@ class TestTypeConversions(unittest.TestCase):
             _python_to_dynamodb, datetime.datetime(2012, 05, 31, 12, 0, 0))
 
 
-class TestGetDefaultValue(unittest.TestCase):
-    @mock.patch("dynamodb_mapper.model.datetime")
-    def test_datetime(self, m_datetime):
-        # Plain default date should be now
-        m_datetime.now.return_value = now = datetime.datetime.now(tz=utc_tz)
-        self.assertEqual(
-            now,
-            _get_default_value(m_datetime))
-        self.assertEqual(utc_tz.tzname(now), "UTC")
-
-        # default date with supplied fallback
-        fallback = datetime.datetime(2012, 05, 31, 12, 0, 0, tzinfo=utc_tz)
-        self.assertEqual(
-            fallback,
-            _get_default_value(datetime.datetime, fallback))
-
-    def test_int(self):
-        # Plain default int
-        self.assertEqual(0, _get_default_value(int))
-        # Default int with fallback
-        self.assertEqual(42, _get_default_value(int, 42))
-        # Default int with callable fallback
-        self.assertEqual(42, _get_default_value(int, return_42))
-
-    def test_incompatible_types_crash_constant(self):
-        # Default unicode with int fallback should fail
-        self.assertRaises(
-            TypeError,
-            _get_default_value,
-            unicode, 42
-        )
-
-    def test_incompatible_types_crash_callable(self):
-        # Default unicode with callable int fallback should fail
-        self.assertRaises(
-            TypeError,
-            _get_default_value,
-            unicode, return_42
-        )
-
-
 class TestDynamoDBModel(unittest.TestCase):
     def setUp(self):
         ConnectionBorg()._connection = None
@@ -640,56 +655,45 @@ class TestDynamoDBModel(unittest.TestCase):
 
     def test_build_default_values(self):
         d = DoomEpisode()
-        self.assertEqual(d.id, 0)
-        self.assertEqual(d.name, u"")
+        self.assertEqual(d.id, None)
+        self.assertEqual(d.name, None)
 
-        # Using to_json_dict because _to_db_dict omits empty values
-        d_dict = d.to_json_dict()
-        self.assertEqual(d_dict["id"], d.id)
-        self.assertEqual(d_dict["name"], d.name)
-
-    @mock.patch("dynamodb_mapper.model._get_default_value")
-    def test_build_default_values_with_defaults(self, m_defaulter):
-        m_defaulter.return_value = u"weak"
+    def test_build_default_values_with_defaults(self):
         d = PlayerStrength()
-
-        m_defaulter.assert_called_with(unicode, u"weak", name='strength')
         self.assertEqual(d.strength, u"weak")
 
+    def test_build_default_values_with_callable_defaults(self):
+        d = SchemaCallableDefault()
+        self.assertEqual(d.id, 42)
+
     def test_init_from_args(self):
         d = DoomEpisode(id=1, name=u"Knee-deep in the Dead")
         self.assertEqual({}, d._raw_data)
         self.assertEqual(1, d.id)
         self.assertEqual(u"Knee-deep in the Dead", d.name)
 
-    def test_init_from_unicode_kwargs(self):
-        # regression test for bug #17
-        kwargs = {
-            u'id': u'1',
-            u'name': u'Knee-deep in the Dead',
-        }
-
-        d = DoomEpisode(**kwargs)
-        self.assertEqual({}, d._raw_data)
-        self.assertEqual(1, d.id)
-        self.assertEqual(u"Knee-deep in the Dead", d.name)
-
-    def test_build_from_dict(self):
-        #FIXME: can it be removed as this feature is implicitly tested elsewhere ?
+    def test_build_from_db_dict(self):
         d_dict = {"id": 1, "name": "Knee-deep in the Dead"}
         d = DoomEpisode._from_db_dict(d_dict)
         self.assertEqual(d_dict, d._raw_data)
         self.assertEqual(d_dict["id"], d.id)
         self.assertEqual(d_dict["name"], d.name)
 
-    def test_build_from_dict_missing_attrs(self):
+    def test_build_from_db_dict_autoinc(self):
+        d_dict = {"id": 1, "text": "toto"}
+        d = LogEntry._from_db_dict(d_dict)
+        self.assertEqual(d_dict, d._raw_data)
+        self.assertEqual(d_dict["id"], autoincrement_int(d.id))
+        self.assertEqual(d_dict["text"], d.text)
+
+    def test_build_from_db_dict_missing_attrs(self):
         #FIXME: can it be removed as this feature is implicitly tested elsewhere ?
         d = DoomCampaign._from_db_dict({})
         self.assertEqual(d.id, 0)
         self.assertEqual(d.name, u"")
         self.assertEqual(d.cheats, set())
 
-    def test_build_from_dict_json_list(self):
+    def test_build_from_db_dict_json_list(self):
         #FIXME: can it be removed as this feature is implicitly tested elsewhere ?
         attacks = [
             {'name': 'fireball', 'damage': 10},
@@ -707,32 +711,6 @@ class TestDynamoDBModel(unittest.TestCase):
         m2 = DoomMonster._from_db_dict({"id": 1})
         self.assertEqual(m2.attacks, [])
 
-    def test_build_from_dict_json_dict(self):
-        #FIXME: can it be removed as this feature is implicitly tested elsewhere ?
-        monsters = {
-            "cacodemon": [
-                {"x": 10, "y": 20},
-                {"x": 10, "y": 30}
-            ],
-            "imp": [],
-            "cyberdemon": []
-        }
-
-        e = DoomMonsterMap._from_db_dict({
-            "map_id": 1,
-            "monsters": json.dumps(monsters, sort_keys=True)
-        })
-        self.assertEqual(e.monsters, monsters)
-
-        e2 = DoomMonsterMap._from_db_dict({"map_id": 1})
-        self.assertEqual(e2.monsters, {})
-
-    def test_build_from_db_dict(self):
-        #FIXME: can it be removed as this feature is implicitly tested elsewhere ?
-        d_dict = {"id": 1, "name": "Knee-deep in the Dead"}
-        d = DoomEpisode._from_db_dict(d_dict)
-        self.assertEqual(d_dict, d._raw_data)
-
     def test_build_from_db_dict_json_dict(self):
         #FIXME: can it be removed as this feature is implicitly tested elsewhere ?
         monsters = {
@@ -754,6 +732,59 @@ class TestDynamoDBModel(unittest.TestCase):
         self.assertEqual(raw_dict["map_id"], e._raw_data["map_id"])
         self.assertEqual(raw_dict["monsters"], e._raw_data["monsters"])
 
+        e2 = DoomMonsterMap._from_db_dict({"map_id": 1})
+        self.assertEqual(e2.monsters, {})
+
+    def test_build_from_db_dict_validators(self):
+        scores = [1,2,3,4]
+
+        # nominal test
+        raw_dict = {
+            "name": u"Jean-Tiare",
+            "age": 22,
+            "scores": json.dumps(scores, sort_keys=True)
+        }
+
+        e = SchemaValidators._from_db_dict(raw_dict)
+        self.assertEqual(u"Jean-Tiare", e.name)
+        self.assertEqual(22, e.age)
+        self.assertEqual(scores, e.scores)
+
+        # empty scores
+        raw_dict = {
+            "name": u"Jean-Tiare",
+            "age": 22,
+        }
+
+        e = SchemaValidators._from_db_dict(raw_dict)
+        self.assertEqual(u"Jean-Tiare", e.name)
+        self.assertEqual(22, e.age)
+        self.assertEqual([], e.scores)
+
+        # bad age
+        raw_dict = {
+            "name": u"Jean-Tiare",
+            "age": -1,
+            "scores": json.dumps(scores, sort_keys=True)
+        }
+        self.assertRaises(Invalid, SchemaValidators._from_db_dict, raw_dict)
+
+        # bad score
+        raw_dict = {
+            "name": u"Jean-Tiare",
+            "age": 22,
+            "scores": json.dumps([101], sort_keys=True)
+        }
+        self.assertRaises(Invalid, SchemaValidators._from_db_dict, raw_dict)
+
+    def test_from_db_dict_validators_bad(self):
+        # unsupported specification in schema
+        raw_dict = {
+            "name": u"Jean-Tiare",
+        }
+
+        self.assertRaises(SchemaError, SchemaValidatorsBad._from_db_dict, raw_dict)
+
     def test_to_db_dict_json_list(self):
         #FIXME: can it be removed as this feature is implicitly tested elsewhere ?
         attacks = [
@@ -790,11 +821,23 @@ class TestDynamoDBModel(unittest.TestCase):
         self.assertEqual(d["map_id"], 1)
         self.assertEqual(d["monsters"], json.dumps(monsters, sort_keys=True))
 
+    def test_to_db_dict_validators(self):
+        e = SchemaValidators()
+        e.name = u"Jean-Tiare"
+        e.age = 22
+        e.scores = [1,2,3,4]
+
+        d = e._to_db_dict()
+
+
+        self.assertEqual(d[u"name"], e.name)
+        self.assertEqual(d[u"age"], e.age)
+        self.assertEqual(d[u"scores"], json.dumps(e.scores))
 
     @mock.patch("dynamodb_mapper.model.Item")
     @mock.patch("dynamodb_mapper.model.boto")
     def test_save_simple_types(self, m_boto, m_item):
-        DoomEpisode().save()
+        DoomEpisode(id=123, name=u"").save()
 
         m_item.return_value.put.assert_called_once()
 
@@ -803,7 +846,7 @@ class TestDynamoDBModel(unittest.TestCase):
     def test_save_sets(self, m_boto, m_item):
         l = DoomCampaign()
         l.id = 1
-        l.name = "Knee-deep in the Dead"
+        l.name = u"Knee-deep in the Dead"
         l.cheats = set(["iddqd", "idkfa", "idclip"])
         l.save()
 
@@ -813,7 +856,7 @@ class TestDynamoDBModel(unittest.TestCase):
     @mock.patch("dynamodb_mapper.model.Item")
     @mock.patch("dynamodb_mapper.model.boto")
     def test_save_empty_sets(self, m_boto, m_item):
-        l = DoomCampaign()
+        l = DoomCampaign(name=u"")
         l.id = 1
         l.cheats = set()
         l.save()
@@ -823,7 +866,7 @@ class TestDynamoDBModel(unittest.TestCase):
     @mock.patch("dynamodb_mapper.model.Item")
     @mock.patch("dynamodb_mapper.model.boto")
     def test_save_empty_strings(self, m_boto, m_item):
-        e = DoomEpisode()
+        e = DoomEpisode(name=u"")
         e.id = 0
         e.save()
 
@@ -879,30 +922,26 @@ class TestDynamoDBModel(unittest.TestCase):
     def test_save_autoincrement_int(self, m_boto, m_item):
         m_item_instance = m_item.return_value
         m_item_instance.hash_key_name = "id"
-        max_item = m_item_instance.table.new_item.return_value
 
-        max_item.save.return_value = {
+        m_item_instance.save.return_value = {
             'Attributes': {
                 '__max_hash_key__': 3
             }
         }
 
         l = LogEntry()
-        l.text = "Everybody's dead, Dave."
+        l.text = u"Everybody's dead, Dave."
         l.save()
 
-        m_item.assert_called_with(mock.ANY, attrs={"id": 0, "text": l.text})
-        m_item.return_value.__setitem__.assert_called_with("id", 3)
-        m_item_instance.save.assert_called()
+        m_item.assert_called_with(mock.ANY, attrs={"id": 3, "text": l.text})
 
     @mock.patch("dynamodb_mapper.model.Item")
     @mock.patch("dynamodb_mapper.model.boto")
     def test_save_autoincrement_int_conflict(self, m_boto, m_item):
         m_item_instance = m_item.return_value
         m_item_instance.hash_key_name = "id"
-        max_item = m_item_instance.table.new_item.return_value
 
-        max_item.save.return_value = {
+        m_item_instance.save.return_value = {
             'Attributes': {
                 '__max_hash_key__': 3
             }
@@ -920,11 +959,10 @@ class TestDynamoDBModel(unittest.TestCase):
         m_item_instance.put.side_effect = err
 
         l = LogEntry()
-        l.text = "Everybody's dead, Dave."
+        l.text = u"Everybody's dead, Dave."
         l.save()
 
-        m_item.return_value.__setitem__.assert_called_with("id", 3)
-        self.assertEqual(m_item.return_value.__setitem__.call_count, 2)
+        self.assertEqual(m_item_instance.save.call_count, 2)
 
     @mock.patch("dynamodb_mapper.model.Item")
     @mock.patch("dynamodb_mapper.model.boto")
@@ -939,7 +977,7 @@ class TestDynamoDBModel(unittest.TestCase):
         m_item_instance.put.side_effect = error
 
         l = LogEntry()
-        l.text = "Everybody's dead, Dave."
+        l.text = u"Everybody's dead, Dave."
         self.assertRaises(MaxRetriesExceededError, l.save)
 
         self.assertEqual(m_item_instance.put.call_count, MAX_RETRIES)
@@ -957,15 +995,15 @@ class TestDynamoDBModel(unittest.TestCase):
         m_item_instance.put.side_effect = error
 
         l = LogEntry()
-        l.text = "Everybody's dead, Dave."
+        l.text = u"Everybody's dead, Dave."
         self.assertRaises(DynamoDBResponseError, l.save)
 
     @mock.patch("dynamodb_mapper.model.Item")
     @mock.patch("dynamodb_mapper.model.boto")
     def test_save_autoincrement_magic_overwrite(self, m_boto, m_item):
         l = LogEntry()
-        l.text = "Everybody's dead, Dave."
-        l.id = MAGIC_KEY
+        l.text = u"Everybody's dead, Dave."
+        l.id = autoincrement_int(MAGIC_KEY)
 
         self.assertRaises(SchemaError, l.save)
 
@@ -977,7 +1015,7 @@ class TestDynamoDBModel(unittest.TestCase):
             None, None, {"__type": "ConditionalCheckFailedException"})
         m_item_instance.put.side_effect = error
 
-        name = "Knee-deep in the Dead"
+        name = u"Knee-deep in the Dead"
         cheats = set(["iddqd", "idkfa"])
 
         c = DoomCampaign()
@@ -1002,7 +1040,7 @@ class TestDynamoDBModel(unittest.TestCase):
             None, None, {"__type": "ConditionalCheckFailedException"})
         m_item_instance.put.side_effect = error
 
-        name = "Knee-deep in the Dead"
+        name = u"Knee-deep in the Dead"
         completed = False
 
         c = DoomCampaignStatus()
@@ -1027,9 +1065,9 @@ class TestDynamoDBModel(unittest.TestCase):
             None, None, {"__type": "ConditionalCheckFailedException"})
         m_item_instance.put.side_effect = error
 
-        c = DoomMap()
+        c = DoomMap(world=u"")
         c.episode_id = 0
-        c.name = "Knee-deep in the Dead"
+        c.name = u"Knee-deep in the Dead"
 
         self.assertRaises(
             OverwriteError,
@@ -1047,7 +1085,8 @@ class TestDynamoDBModel(unittest.TestCase):
 
         c = DoomMap()
         c.episode_id = 0
-        c.name = "Knee-deep in the Dead"
+        c.name = u"Knee-deep in the Dead"
+        c.world = u"Oui-Oui"
 
         self.assertRaises(DynamoDBResponseError, c.save)
 
@@ -1076,8 +1115,8 @@ class TestDynamoDBModel(unittest.TestCase):
         self.assertEqual(DATA[1][u'id'], res[1].id)
         self.assertEqual(DATA[1][u'name'], res[1].name)
 
-    @mock.patch("dynamodb_mapper.model.boto")
-    def test_get_batch_hash_range_key(self, m_boto):
+    @mock.patch("dynamodb_mapper.model.ConnectionBorg")
+    def test_get_batch_hash_range_key(self, m_borg):
         KEYS = [(1, u"The level with cute cats"),
                 (2, u"The level with horrible monsters")]
         DATA = [{
@@ -1091,7 +1130,7 @@ class TestDynamoDBModel(unittest.TestCase):
                     u"world": u"Hell",
                 }]
 
-        m_table = m_boto.connect_dynamodb.return_value.get_table.return_value
+        m_table = m_borg.return_value.get_table.return_value
         m_table.batch_get_item.return_value = DATA
 
         res = DoomMap.get_batch(KEYS)
